from django.contrib.auth.models import User
from rest_framework import serializers

from dra.serializers import ActionSerializer

class UserSerializer(ActionSerializer,
                     serializers.ModelSerializer):

    def get_fullname(self, obj):
        return '{} {}'.format(obj.first_name, obj.last_name)

    class Meta:
        model = User
        fields = (
            'username', 'email',
            'first_name', 'last_name'
        )
        action_fields_map = {
            'retrieve': {
                'fields': fields,
                'exclude': ('username',),
                'custom_fields': {
                    'fullname': serializers.SerializerMethodField(read_only=True)
                }
            }
        }

class NoActionConfigUserSerializer(UserSerializer, ActionSerializer):

    class Meta:
        model = User
        fields = (
            'username', 'email',
            'first_name', 'last_name'
        )


from django.contrib.auth.models import User
from django.shortcuts import reverse

from rest_framework.test import APITestCase
from rest_framework.response import Response

from .serializers import UserSerializer, NoActionConfigUserSerializer

class DRATestCase(APITestCase):
    
    def setUp(self):
        self.user = User.objects.create(
            username='myuser',
            email='myuser@email.com',
            first_name='my',
            last_name='user',
            password='mypassword'
        )

    def test_list_action_not_retrieve_fullname_attribute(self):
        """
        This attribute is configured to display only in retrieve action
        on UserSerializer
        """
        response: Response = self.client.get(reverse('user-list'))
        
        user_data = response.data[0]

        self.assertNotIn('fullname', user_data.keys())

    def test_list_action_retrieves_username_attribute(self):
        """
        This attribute is configured to display only in list action
        on UserSerializer
        """
        response: Response = self.client.get(reverse('user-list'))
        
        user_data = response.data[0]

        self.assertIn('username', user_data.keys())

    def test_list_action_not_retrieves_fullname_attribute(self):
        """
        fullname is a SerializerMethodField configured to display only in retrive action
        """
        response: Response = self.client.get(reverse('user-list'))
        
        user_data = response.data[0]

        self.assertNotIn('fullname', user_data.keys())

    def test_retrieve_action_displays_fullname_attribute_and_no_username(self):
        """
        This attribute is configured to display only in list action
        on UserSerializer
        """
        response: Response = self.client.get(
            reverse(
                'user-detail',
                kwargs={'pk': self.user.id}
            )
        )
        
        user_data = response.data

        self.assertIn('fullname', user_data.keys())
        self.assertNotIn('username', user_data.keys())

    def test_no_view_passed_to_serializer(self):
        """
        When using the serializer manually, it is not possible
        to determine the action (list, retrive), so the default
        fields are always returned
        """
        serializer_instance = UserSerializer(self.user)
        serializer_data = serializer_instance.data

        self.assertIn('username', serializer_data)
    
    def test_no_action_fields_map_provided(self):
        """
        When the attribute `action_fields_map` is not provided,
        the default fields are always returned
        """
        response: Response = self.client.get(
            reverse(
                'user-no-action-config-detail',
                kwargs={'pk': self.user.id}
            )
        )

        self.assertIn('username', response.data.keys())


from rest_framework import serializers

class ActionSerializer(serializers.Serializer):
    """
    Mixin to add action_fields_map attribute to serializer's Meta.
    ex: 
        class Meta:
            fields = ('url', 'user',)
            action_fields_map = {
                'retrieve': {
                    'fields': fields + ('checked',)
                    'exclude_fields': ['url'],
                    'custom_fields': {
                        'stories': StorySerializer(read_only=True, many=True)
                    }
                }
            }
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        meta = getattr(self, 'Meta', None)
        # attribute to configure the action fields mapping
        self.action_fields_map: dict = getattr(meta, 'action_fields_map', None)
        self._validate_action_fields_map(self.action_fields_map)

    def _validate_action_fields_map(self, action_fields_map: dict):
        if action_fields_map:
            # Asserting action_fields_map is a dictionary.
            assert isinstance(action_fields_map, dict), \
                'The attribute action_fields_map must be instance of dict'
            
            # Iterating over the actions
            for action, config in action_fields_map.items():
                assert isinstance(action, str), \
                    'The action name must be a string not {}'.format(type(action).__name__)

                assert isinstance(config, dict), \
                    'The configuration for `{}` action must be a dictionary not {}'.format(action, type(config).__name__)

                fields = config.get('fields', None)
                exclude = config.get('exclude', None)
                custom_fields = config.get('custom_fields', None)
                
                if fields is not None:
                    assert isinstance(fields, (list, tuple)), \
                        'The `fields` must be a list. Got `{}`'.format(type(fields).__name__)

                if exclude is not None:
                    assert isinstance(exclude, (list, tuple)), \
                        '`exclude` attribute must be a list'
                
                if custom_fields is not None:
                    assert isinstance(custom_fields, dict), \
                        '`custom_fields` must be a dictionary'

    def _get_action_config(self, current_action: str, action_fields_map: dict) -> dict:
        return action_fields_map.get(current_action, None)

    # Overriding to include the new fields or to remove fields
    def get_field_names(self, declared_fields, info):
        # Checking if the view is passed to the serializer instance
        if not (self.context and self.context.get('view', None)):
            return super().get_field_names(declared_fields, info)

        if self.action_fields_map is None:
            return super().get_field_names(declared_fields, info)
        
        current_action = self.context.get('view').action
        current_action_config = self._get_action_config(
            current_action,
            self.action_fields_map
        )

        if current_action_config is None:
            return super().get_field_names(declared_fields, info)
        
        fields = list(current_action_config.get('fields', None))
        fields_to_remove = current_action_config.get('exclude', None)

        if fields_to_remove:
            for field in fields_to_remove:
                fields.remove(field)
        
        return fields

    # Overriding the default get_field method
    def get_fields(self):
        declared_fields = super().get_fields()
        # Checking if the view is passed to the serializer instance
        if not (self.context and self.context.get('view', None)):
            return declared_fields

        if self.action_fields_map is None:
            return declared_fields

        current_action = self.context.get('view').action

        current_action_config = self._get_action_config(
            current_action,
            self.action_fields_map
        )

        if not current_action_config:
            return declared_fields
                
        new_fields = declared_fields.copy()

        custom_fields: dict = current_action_config.get('custom_fields')
        if custom_fields:
            for key, value in custom_fields.items():
                new_fields[key] = value
    
        return new_fields



from rest_framework.routers import DefaultRouter

from .viewsets import UserModelViewSet, UserNoActionConfigModelViewSet

router = DefaultRouter()
router.register(r'users', UserModelViewSet, basename='user')
router.register(
    r'users-no-action-config',
    UserNoActionConfigModelViewSet,
    basename='user-no-action-config'
)

urlpatterns = router.urls


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='django-rest-action-serializer',
    version='1.0.0',
    packages=setuptools.find_packages(),
    description='A Django app that provides a serializer that allows You to customize the fields'
                'according to the action provided without the need to'
                'create other serializers.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Bruno Barreto Freitas',
    author_email='brunobarretofreitas@outlook.com',
    url='https://github.com/brunobarretofreitas/django-rest-action-serializer',
    download_url='https://github.com/brunobarretofreitas/django-rest-action-serializer/archive/master.zip',
    keywords='django django-rest-framework api serializer',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=[
        'Django', 'djangorestframework'
    ]
)


from django.contrib.auth import models
from rest_framework.viewsets import ModelViewSet

from .serializers import UserSerializer, NoActionConfigUserSerializer

class UserModelViewSet(ModelViewSet):
    queryset = models.User.objects.all()
    serializer_class = UserSerializer

class UserNoActionConfigModelViewSet(UserModelViewSet):
    serializer_class = NoActionConfigUserSerializer